package ch.sahits.game.graphic.image;

import java.awt.Font;
import java.awt.FontFormatException;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.font.GlyphVector;
import java.awt.image.BufferedImage;
import java.io.IOException;

public interface IOpenPatricianPainter {

	/**
	 * Create a BufferedImage of the specified dimensions. The image is tiled with <code>imageName</code>.
	 * @param imageName of the image
	 * @param width of the image
	 * @param heigth of the image
	 * @param comp Component on which is drawn
	 * @return
	 */
	public abstract BufferedImage createTiledImage(String imageName, int width,
			int heigth);

	/**
	 * Paint the image onto the component. If the image is larger the image is clipped. If the image
	 * is smaller it is repeated.
	 * @param g Graphics used to draw on
	 * @param imageName name of the tiled image
	 * @param comp Component on which to draw
	 * @param rect rectangle which is to be drawn of the whole
	 */
	public abstract void drawBackground(Graphics g, String imageName,
			Rectangle rect);

	/**
	 * Wrapper method for drawing a tile at a specified position
	 * @param img Original image to be used
	 * @param x position of the top right corner
	 * @param y position of the top right corner
	 * @param width of the image to be drawn (the rest is clipped away)
	 * @param height of the image to be drawn (the rest is clipped away)
	 * @param comp Component on which the image is drawn
	 * @param g Graphics used for the drawing
	 */
	public abstract void drawTile(BufferedImage img, int x, int y, int width,
			int height, Graphics g);

	/**
	 * Draw a string on the component
	 * @param graphics Graphics context that is used
	 * @param s String to be drawn
	 * @param fontSize Font size in pt to be used
	 * @param posX Center X position of the string
	 * @param posY Center y position of the string
	 */
	public abstract void drawString(Graphics2D graphics, String s,
			int fontSize, int posX, int posY);

	/**
	 * Create the glyph vector using the string
	 * @param graphics to be used
	 * @param s String
	 * @param fontSize font size in points
	 * @return {@link GlyphVector}
	 * @throws FontFormatException
	 * @throws IOException
	 */
	public abstract GlyphVector createGlyphVector(Graphics2D graphics,
			String s, int fontSize) throws FontFormatException, IOException;

	/**
	 * Create a glyph vector based on a font
	 * @param graphics
	 * @param s
	 * @param font
	 * @return
	 */
	public abstract GlyphVector createGlyphVector(Graphics2D graphics,
			String s, Font font);

	/**
	 * Create the glyph vector using the string
	 * @param s String
	 * @param fontSize font size in points
	 * @return {@link GlyphVector}
	 * @throws FontFormatException
	 * @throws IOException
	 */
	public abstract GlyphVector createGlyphVector(String s, int fontSize)
			throws FontFormatException, IOException;

	/**
	 * Create turned steering wheel. 0% is whole on the left and 100% whole on the right
	 * @param width width of the component
	 * @param percent of the turning
	 */
	public abstract BufferedImage createSteeringWheel(int width, float percent);

	/**
	 * Create the polygon that represents the handle of the steering wheel
	 * @param width of the component the polygon is based upon
	 * @param percent of the rotation
	 * @return
	 */
	public abstract Polygon getSteeringWheelHandle(int width, float percent);

	/**
	 * Compute the slider track rectangle.
	 * @param orientation
	 * @param width
	 * @param castMin cast value of the minimum
	 * @param castMax cast value of the maximum
	 * @return
	 */
	public abstract Rectangle computeSliderTrack(int width, float castMin,
			float castMax);

	/**
	 * Scale an input image along the x-axis only. The scaled dimension may not exactly match the
	 * desired, but it will be best match
	 * @param img Image to be scaled
	 * @param destWidth width of the scaled image
	 * @return scaled image or null if the input was null
	 */
	public abstract BufferedImage scaleX(BufferedImage img, int destWidth);

	/**
	 * Scale an input image along the y-axis only. The scaled dimension may not exactly match the
	 * desired, but it will be best match
	 * @param img Image to be scaled
	 * @param destHeight height of the scaled image
	 * @return scaled image or null if the input was null
	 */
	public BufferedImage scaleY(BufferedImage img, int destHeight);

}