package ch.sahits.game.graphic.image.impl;

import java.awt.Font;
import java.awt.FontFormatException;
import java.io.IOException;
import java.util.HashMap;

import ch.sahits.game.graphic.image.IFontLoader;
/**
 * Font loader for loading fonts.
 * This is implemented as a singleton
 * @author Andi Hotz, (c) Sahits GmbH, 2011
 * Created on Nov 13, 2011
 *
 */
public class FontLoader implements IFontLoader {
	
	private static FontLoader INSTANCE = null;
	private String frakturFontName ="Goeschen.ttf";
	private String decoFontName ="RothenburgDecorative.ttf";
	private String celtDecoFontName = "CELTICMD.TTF";
	private String defaultFont = "starnberg.ttf";
	private String handwritten = "DobkinScript.ttf";
	private String serif = "WeekdaysRomanSlant.ttf";
	/** Cash for already loaded fonts */
	private HashMap<String, Font> fontCash = new HashMap<String, Font>();
	/** Hide default constructor */
	private FontLoader(){}
	/**
	 * Retrieve the singleton instance.
	 * @return
	 */
	public static FontLoader getInstance(){
		if (INSTANCE==null){
			INSTANCE=new FontLoader();
		}
		return INSTANCE;
	}
	
	/* (non-Javadoc)
	 * @see ch.sahits.game.graphic.image.impl.IFontLoader#createGothicFont()
	 */
	@Override
	public Font createGothicFont() throws FontFormatException, IOException{
		Font font = Font.createFont( Font.TRUETYPE_FONT, 
                getClass().getResourceAsStream( "/"+frakturFontName) );
		return font;
	}
	/**
	 * Retrieve the default decorative font
	 * @return
	 * @throws FontFormatException
	 * @throws IOException
	 */
	@Override
	public Font createDefaultDecorativeFont() throws FontFormatException, IOException{
		return createDecorative1Font();
	}
	/**
	 * Load a decorative font with baroque capitals.
	 * @return
	 * @throws FontFormatException
	 * @throws IOException
	 */
	public Font createDecorative1Font() throws FontFormatException, IOException{
		return getFont(decoFontName);
	}
	/**
	 * Retrieve the font
	 * @param fontName Name of the font
	 * @return
	 * @throws FontFormatException
	 * @throws IOException
	 */
	private Font getFont(String fontName) throws FontFormatException, IOException{
		return getFont(fontName,Font.TRUETYPE_FONT);
	}
	/**
	 * Retrieve the font with a certain type
	 * @param fontName Name of the Font
	 * @param fontType type of the font as defined in Font
	 * @return
	 * @throws FontFormatException
	 * @throws IOException
	 */
	private Font getFont(String fontName, int fontType) throws FontFormatException, IOException{
		Font f = fontCash.get(fontName);
		if (f==null){
			f = Font.createFont( fontType, 
	                getClass().getResourceAsStream( "/"+fontName) );
			fontCash.put(fontName, f);
		}
		return f;
	}
	/**
	 * Load the default decorative font with celtic capitals
	 * @return
	 * @throws FontFormatException
	 * @throws IOException
	 */
	public Font createDecorative2Font()throws FontFormatException, IOException{
		return getFont(celtDecoFontName);
	}
	/* (non-Javadoc)
	 * @see ch.sahits.game.graphic.image.impl.IFontLoader#createStarnbergFont()
	 */
	@Override
	public Font createStarnbergFont()throws FontFormatException, IOException{
		return getFont(defaultFont);
	}
	/* (non-Javadoc)
	 * @see ch.sahits.game.graphic.image.impl.IFontLoader#createDefaultFont()
	 */
	@Override
	public Font createDefaultFont()throws FontFormatException, IOException{
		return getFont(serif);
	}
	/* (non-Javadoc)
	 * @see ch.sahits.game.graphic.image.impl.IFontLoader#createDefaultFont(int)
	 */
	@Override
	public Font createDefaultFont(int size)throws FontFormatException, IOException{
		Font font = createDefaultFont();
		font = font.deriveFont((float)size);
		return font;
	}
	/* (non-Javadoc)
	 * @see ch.sahits.game.graphic.image.impl.IFontLoader#createStranbergFont(int)
	 */
	@Override
	public Font createStranbergFont(int size)throws FontFormatException, IOException{
		Font font = createStarnbergFont();
		font = font.deriveFont((float)size);
		return font;
	}
	/* (non-Javadoc)
	 * @see ch.sahits.game.graphic.image.impl.IFontLoader#createDefaultHandwrittenFont()
	 */
	@Override
	public Font createDefaultHandwrittenFont()throws FontFormatException, IOException{
		return getFont(handwritten);
	}

}
